from django.apps import apps

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Schedule periodic tasks'

    def handle(self, *args, **options):
        apps.get_app_config('django_cloud_tasks').schedule_tasks()
        self.stdout.write(self.style.SUCCESS('Successfully scheduled tasks'))



# pylint: disable=not-callable, unsubscriptable-object
import abc
import logging
import os
from datetime import datetime, timedelta

from google import auth
from google.api_core.exceptions import NotFound
from google.cloud import tasks_v2, scheduler
from google.protobuf import timestamp_pb2
from googleapiclient.discovery import build

logger = logging.getLogger(__name__)

DEFAULT_LOCATION = 'us-east1'
DEFAULT_TIMEZONE = 'UTC'
DEFAULT_PROJECT_ID = os.environ.get('PROJECT_ID')


class GoogleCloudPilotAPI(abc.ABC):
    _client_class = None
    _scopes = ['https://www.googleapis.com/auth/cloud-platform']

    def __init__(self, subject=None, **kwargs):
        self.credentials, project_id = self._build_credentials(subject=subject)

        self.project_id = kwargs.get('project') \
            or DEFAULT_PROJECT_ID \
            or getattr(self.credentials, 'project_id') \
            or project_id

        self.client = (self._client_class or build)(
            credentials=self.credentials,
            **kwargs
        )

    @classmethod
    def _build_credentials(cls, subject=None):
        credentials, project_id = auth.default()

        if subject:
            credentials = credentials.with_subject(subject=subject)

        return credentials, project_id

    @property
    def oidc_token(self):
        return {'oidc_token': {'service_account_email': self.credentials.service_account_email}}


class CloudTasksClient(GoogleCloudPilotAPI):
    _client_class = tasks_v2.CloudTasksClient
    DEFAULT_METHOD = tasks_v2.HttpMethod.POST

    def push(self, name, queue, url, payload, method=DEFAULT_METHOD, delay_in_seconds=0):
        parent = self.client.queue_path(self.project_id, queue)

        tasks_v2.Task(
            name=name,
            http_request=tasks_v2.HttpRequest(
                http_method=method,
                url=url,
                body=payload.encode(),
                **self.oidc_token,
            )
        )
        task = {
            'http_request': {
                'http_method': method,
                'url': url,
                'body': payload.encode(),
                **self.oidc_token,
            },
            'name': name,
        }

        if delay_in_seconds:
            target_date = datetime.utcnow() + timedelta(seconds=delay_in_seconds)
            timestamp = timestamp_pb2.Timestamp()
            timestamp.FromDatetime(target_date)

            task.schedule_time = timestamp

        response = self.client.create_task(parent, task)
        return response


class CloudSchedulerClient(GoogleCloudPilotAPI):
    _client_class = scheduler.CloudSchedulerClient
    DEFAULT_METHOD = scheduler.HttpMethod.POST

    def create(
            self,
            name,
            url,
            payload,
            cron,
            timezone=DEFAULT_TIMEZONE,
            method=DEFAULT_METHOD,
            headers=None,
            location=DEFAULT_LOCATION,
    ):
        parent = f"projects/{self.project_id}/locations/{location}"
        job = scheduler.Job(
            name=f'{parent}/jobs/{name}',
            schedule=cron,
            time_zone=timezone,
            http_target=scheduler.HttpTarget(
                uri=url,
                http_method=method,
                body=payload.encode(),
                headers=headers or {},
            )
        )

        response = self.client.create_job(
            request={
                "parent": parent,
                "job": job
            }
        )
        return response

    def update(
            self,
            name,
            url,
            payload,
            cron,
            timezone=DEFAULT_TIMEZONE,
            method=DEFAULT_METHOD,
            headers=None,
            location=DEFAULT_LOCATION,
    ):
        parent = f"projects/{self.project_id}/locations/{location}"
        job = scheduler.Job(
            name=f'{parent}/jobs/{name}',
            schedule=cron,
            time_zone=timezone,
            http_target=scheduler.HttpTarget(
                uri=url,
                http_method=method,
                body=payload.encode(),
                headers=headers or {},
            )
        )

        response = self.client.update_job(
            job=job,
        )
        return response

    def get(self, name, location=DEFAULT_LOCATION):
        response = self.client.get_job(
            name=f"projects/{self.project_id}/locations/{location}/jobs/{name}",
        )
        return response

    def delete(self, name, location=DEFAULT_LOCATION):
        response = self.client.delete_job(
            name=f"projects/{self.project_id}/locations/{location}/jobs/{name}",
        )
        return response

    def put(
            self,
            name,
            url,
            payload,
            cron,
            timezone=DEFAULT_TIMEZONE,
            method=DEFAULT_METHOD,
            headers=None,
            location=DEFAULT_LOCATION,
    ):
        try:
            response = self.update(
                name=name,
                url=url,
                payload=payload,
                cron=cron,
                timezone=timezone,
                method=method,
                headers=headers,
                location=location,
            )
        except NotFound:
            response = self.create(
                name=name,
                url=url,
                payload=payload,
                cron=cron,
                timezone=timezone,
                method=method,
                headers=headers,
                location=location,
            )
        return response



import json

from django.apps import apps
from django.http import HttpResponse
from django.views.generic import View


class GoogleCloudTaskView(View):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.tasks = apps.get_app_config('django_cloud_tasks').tasks

    def post(self, request, task_name, *args, **kwargs):
        try:
            task_class = self.tasks[task_name]
            data = json.loads(request.body)
            output, success = task_class().execute(data=data)
            if success:
                status = 200
                result = {'result': output}
            else:
                status = 400
                result = {'error': output}
        except KeyError:
            status = 404
            result = {'error': f"Task {task_name} not found"}

        response = HttpResponse(status=status, content=json.dumps(result), content_type='application/json')
        return response



import json
from abc import ABC, abstractmethod

from django.apps import apps
from django.urls import reverse

from django_cloud_tasks.client import CloudTasksClient, CloudSchedulerClient


class Task(ABC):
    @abstractmethod
    def run(self, **kwargs):
        raise NotImplementedError()

    def execute(self, data):
        try:
            output = self.run(**data)
            success = True
        except Exception as e:  # pylint: disable=broad-except
            output = str(e)
            success = False
        return output, success

    def delay(self, **kwargs):
        payload = kwargs

        return self.__client.push(
            name=self.name(),
            queue=self.queue,
            url=self.url(),
            payload=json.dumps(payload),
        )

    @classmethod
    def name(cls):
        return cls.__name__

    @property
    def queue(self):
        return 'tasks'

    @classmethod
    def url(cls):
        domain = apps.get_app_config('django_cloud_tasks').domain
        path = reverse('tasks-endpoint', args=(cls.name(),))
        return f'{domain}{path}'

    @property
    def __client(self):
        return CloudTasksClient()


class PeriodicTask(Task, ABC):
    run_every = None

    def delay(self, **kwargs):
        payload = kwargs

        return self.__client.put(
            name=self.schedule_name,
            url=self.url(),
            payload=json.dumps(payload),
            cron=self.run_every,
        )

    @property
    def schedule_name(self):
        return self.name()

    @property
    def __client(self):
        return CloudSchedulerClient()



import abc
import importlib.util
import inspect
import os

from django.apps import AppConfig, apps
from django.conf import settings


class DjangoCloudTasksAppConfig(AppConfig):
    name = 'django_cloud_tasks'
    verbose_name = "Django Cloud Tasks"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tasks = {}
        self.domain = self._fetch_config(name='DOMAIN_URL', default='http://localhost:8080')
        self.location = self._fetch_config(name='GOOGLE_CLOUD_LOCATION', default='us-east1')

    def _fetch_config(self, name, default):
        return getattr(settings, name, os.environ.get(name, default))

    def ready(self):
        self.register_tasks()

    def register_tasks(self):
        for app in apps.get_app_configs():
            if app.name.startswith('django.') or app.name == 'django_cloud_tasks':
                continue
            for task_class in self._scan_tasks(app=app):
                self._register_task(task_class=task_class)
        return {}

    def _scan_tasks(self, app):
        from django_cloud_tasks.tasks import Task  # pylint: disable=import-outside-toplevel

        files = []

        tasks_file = os.path.join(app.path, 'tasks.py')
        tasks_package = os.path.join(app.path, 'tasks')

        if os.path.exists(tasks_file):
            files.append(tasks_file)
        elif os.path.exists(tasks_package):
            files = []
            for path, _, located_files in os.walk(tasks_package):
                files.extend([
                    os.path.join(path, located_file)
                    for located_file in located_files
                    if located_file != '__init__.py'
                ])

        for file in files:
            name = file.rsplit('/', 1)[0].replace('.py', '')
            spec = importlib.util.spec_from_file_location(name, file)
            foo = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(foo)
            for _, klass in inspect.getmembers(foo):
                is_task = inspect.isclass(klass) and issubclass(klass, Task)
                if not is_task:
                    continue

                is_abstract = abc.ABC in klass.__bases__
                if not is_abstract:
                    yield klass

    def _register_task(self, task_class):
        self.tasks[task_class.name()] = task_class

    def schedule_tasks(self):
        for task_klass in self.tasks.values():
            from django_cloud_tasks.tasks import PeriodicTask  # pylint: disable=import-outside-toplevel

            if issubclass(task_klass, PeriodicTask):
                task_klass().delay()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_cloud_tasks',
 'django_cloud_tasks.management',
 'django_cloud_tasks.management.commands']

package_data = \
{'': ['*']}

install_requires = \
['django>=3,<4',
 'google-api-python-client',
 'google-cloud-scheduler>=2,<3',
 'google-cloud-tasks']

setup_kwargs = {
    'name': 'django-google-cloud-tasks',
    'version': '0.0.1rc3',
    'description': 'Async Tasks with HTTP endpoints',
    'long_description': None,
    'author': 'Joao Daher',
    'author_email': 'joao@daher.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
